# TODO: LinearAlgebra.qr! method that works with any workspace
import LinearAlgebra.LAPACK: geqrf!, ormqr!, geqrt!, geqp3!

"""
    QRWs

Workspace for standard [`LinearAlgebra.QR`](https://docs.julialang.org/en/v1/stdlib/LinearAlgebra/#LinearAlgebra.QR)
factorization using the [`LAPACK.geqrf!`](@ref) function.

# Examples
```jldoctest
julia> A = [1.2 2.3
            6.2 3.3]
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3

julia> ws = QRWs(A)
QRWs{Float64}
  work: 64-element Vector{Float64}
  τ: 2-element Vector{Float64}

julia> t = QR(LAPACK.geqrf!(ws, A)...)
QR{Float64, Matrix{Float64}, Vector{Float64}}
Q factor:
2×2 QRPackedQ{Float64, Matrix{Float64}, Vector{Float64}}:
 -0.190022  -0.98178
 -0.98178    0.190022
R factor:
2×2 Matrix{Float64}:
 -6.31506  -3.67692
  0.0      -1.63102

julia> Matrix(t)
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3
```
"""
struct QRWs{T<:Number} <: Workspace
    work::Vector{T}
    τ::Vector{T}
end

Base.length(ws::QRWs) = length(ws.τ)

for (geqrf, elty) in ((:dgeqrf_, :Float64),
                      (:sgeqrf_, :Float32),
                      (:zgeqrf_, :ComplexF64),
                      (:cgeqrf_, :ComplexF32))
    @eval begin
        function Base.resize!(ws::QRWs, A::StridedMatrix{$elty})
            m, n = size(A)
            lda = max(1, stride(A, 2))
            resize!(ws.τ, min(m, n))
            info = Ref{BlasInt}()
            ccall((@blasfunc($geqrf), liblapack), Cvoid,
                  (Ref{BlasInt}, Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt}, Ptr{$elty},
                   Ptr{$elty}, Ref{BlasInt}, Ptr{BlasInt}),
                  m, n, A, lda, ws.τ, ws.work, -1, info)
            chklapackerror(info[])
            resize!(ws.work, BlasInt(real(ws.work[1])))
            return ws
        end
        QRWs(A::StridedMatrix{$elty}) =
            resize!(QRWs(Vector{$elty}(undef, 1), Vector{$elty}(undef, 1)), A)

        function geqrf!(ws::QRWs, A::AbstractMatrix{$elty}; resize=true)
            require_one_based_indexing(A)
            chkstride1(A)
            m, n = size(A)
            if length(ws) < min(m, n)
                if resize
                    resize!(ws, A)
                else
                    throw(ArgumentError("Workspace is too small, use resize!(ws, A)."))
                end
            end
            lda = max(1, stride(A, 2))
            lwork = length(ws.work)
            info = Ref{BlasInt}() # This actually doesn't cause allocations
            ccall((@blasfunc($geqrf), liblapack), Cvoid,
                  (Ref{BlasInt}, Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt}, Ptr{$elty},
                   Ptr{$elty}, Ref{BlasInt}, Ptr{BlasInt}),
                  m, n, A, lda, ws.τ, ws.work, lwork, info)
            chklapackerror(info[])
            return A, ws.τ
        end
    end
end

for (ormqr, elty) in ((:dormqr_, :Float64),
                      (:sormqr_, :Float32))
    @eval begin
        function ormqr!(ws::QRWs{$elty}, side::AbstractChar, trans::AbstractChar,
                        A::AbstractMatrix{$elty},
                        C::AbstractVecOrMat{$elty}; resize=true)
            require_one_based_indexing(A, C)
            chktrans(trans)
            chkside(side)
            chkstride1(A, C)
            m, n = ndims(C) == 2 ? size(C) : (size(C, 1), 1)
            mA   = size(A, 1)
            k    = length(ws.τ)
            if side == 'L' && m != mA
                throw(DimensionMismatch("for a left-sided multiplication, the first dimension of C, $m, must equal the second dimension of A, $mA"))
            end
            if side == 'R' && n != mA
                throw(DimensionMismatch("for a right-sided multiplication, the second dimension of C, $m, must equal the second dimension of A, $mA"))
            end
            if side == 'L' && k > m
                throw(DimensionMismatch("invalid number of reflectors: k = $k should be <= m = $m"))
            end
            if side == 'R' && k > n
                throw(DimensionMismatch("invalid number of reflectors: k = $k should be <= n = $n"))
            end
            info = Ref{BlasInt}()
            ccall((@blasfunc($ormqr), liblapack), Cvoid,
                  (Ref{UInt8}, Ref{UInt8}, Ref{BlasInt}, Ref{BlasInt},
                   Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt}, Ptr{$elty},
                   Ptr{$elty}, Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt},
                   Ref{BlasInt}, Clong, Clong),
                  side, trans, m, n,
                  k, A, max(1, stride(A, 2)), ws.τ,
                  C, max(1, stride(C, 2)), ws.work, length(ws.work),
                  info, 1, 1)
            chklapackerror(info[])
            return C
        end
    end

    for elty2 in (eval(:(Transpose{$elty,<:StridedMatrix{$elty}})),
                  eval(:(Adjoint{$elty,<:StridedMatrix{$elty}})))
        @eval begin
            function ormqr!(ws::QRWs{$elty}, side::AbstractChar, trans::AbstractChar,
                            A::$elty2,
                            C::StridedMatrix{$elty})
                chktrans(trans)
                chkside(side)
                trans = trans == 'T' ? 'N' : 'T'
                return LAPACK.ormqr!(ws, side, trans, A.parent, C)
            end
        end
    end
end

"""
    geqrf!(ws, A; resize=true) -> (A, ws.τ)

Compute the `QR` factorization of `A`, `A = QR`, using previously allocated [`QRWs`](@ref) workspace `ws`.
`ws.τ` contains scalars which parameterize the elementary reflectors of the factorization.
`ws.τ` must have length greater than or equal to the smallest dimension of `A`.
If this is not the case, and `resize==true` the workspace will be automatically
resized to the appropriate size.

`A` and `ws.τ` modified in-place.
"""
geqrf!(ws::QRWs, A::AbstractMatrix; kwargs...)

"""
    ormqr!(ws, side, trans, A, C) -> C

Computes `Q * C` (`trans = N`), `transpose(Q) * C` (`trans = T`), `adjoint(Q) * C`
(`trans = C`) for `side = L` or the equivalent right-sided multiplication
for `side = R` using `Q` from a `QR` factorization of `A` computed using
`geqrf!`.
Uses preallocated workspace `ws` and the factors are assumed to be stored in `ws.τ`.
`C` is overwritten.
"""
ormqr!(ws::QRWs, side::AbstractChar, trans::AbstractChar, A::AbstractMatrix,
       C::AbstractVecOrMat)

"""
    QRWYWs

Workspace to be used with the [`LinearAlgebra.QRCompactWY`](https://docs.julialang.org/en/v1/stdlib/LinearAlgebra/#LinearAlgebra.QRCompactWY)
representation of the blocked QR factorization which uses the [`LAPACK.geqrt!`](@ref) function.
By default the blocksize for the algorithm is taken as `min(36, min(size(template)))`, this can be
overridden by using the `blocksize` keyword of the constructor.

# Examples
```jldoctest
julia> A = [1.2 2.3
            6.2 3.3]
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3

julia> ws = QRWYWs(A)
QRWYWs{Float64, Matrix{Float64}}
  work: 4-element Vector{Float64}
  T: 2×2 Matrix{Float64}

julia> t = QRCompactWY(LAPACK.geqrt!(ws, A)...)
QRCompactWY{Float64, Matrix{Float64}, Matrix{Float64}}
Q factor:
2×2 QRCompactWYQ{Float64, Matrix{Float64}, Matrix{Float64}}:
 -0.190022  -0.98178
 -0.98178    0.190022
R factor:
2×2 Matrix{Float64}:
 -6.31506  -3.67692
  0.0      -1.63102

julia> Matrix(t)
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3
```
"""
mutable struct QRWYWs{R<:Number,MT<:StridedMatrix{R}} <: Workspace
    work::Vector{R}
    T::MT
end

for (geqrt, elty) in ((:dgeqrt_, :Float64),
                      (:sgeqrt_, :Float32),
                      (:zgeqrt_, :ComplexF64),
                      (:cgeqrt_, :ComplexF32))
    @eval begin
        function Base.resize!(ws::QRWYWs, A::StridedMatrix{$elty}; blocksize=36)
            require_one_based_indexing(A)
            chkstride1(A)
            m, n = BlasInt.(size(A))
            @assert n > 0 ArgumentError("Not a Matrix")
            m1 = min(m, n)
            nb = min(m1, blocksize)
            ws.T = similar(ws.T,  nb, m1)
            resize!(ws.work, nb*n)
            return ws
        end
        QRWYWs(A::StridedMatrix{$elty}; kwargs...) = 
            resize!(QRWYWs($elty[], Matrix{$elty}(undef, 0, 0)), A; kwargs...)

        function geqrt!(ws::QRWYWs, A::AbstractMatrix{$elty}; resize=true)
            require_one_based_indexing(A)
            chkstride1(A)
            m, n = size(A)
            minmn = min(m, n)
            nb = size(ws.T, 1)
            if nb > minmn
                if resize
                    resize!(ws, A)
                    nb = size(ws.T, 1)
                else
                    throw(ArgumentError("Allocated workspace block size $nb > $minmn too large.\nUse resize!(ws, A)."))
                end
            end
            
            lda = max(1, stride(A, 2))
            work = ws.work
            info = Ref{BlasInt}()
            ccall((@blasfunc($geqrt), liblapack), Cvoid,
                  (Ref{BlasInt}, Ref{BlasInt}, Ref{BlasInt}, Ptr{$elty},
                   Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt}, Ptr{$elty},
                   Ptr{BlasInt}),
                  m, n, nb, A,
                  lda, ws.T, max(1, stride(ws.T, 2)), ws.work,
                  info)
            chklapackerror(info[])
            return A, ws.T
        end
    end
end

"""
    geqrt!(ws, A; resize=true) -> (A, ws.T)

Compute the blocked `QR` factorization of `A`, `A = QR`, using a preallocated [`QRWYWs`](@ref) workspace `ws`. `ws.T` contains upper
triangular block reflectors which parameterize the elementary reflectors of
the factorization. The first dimension of `ws.T` sets the block size and it must
satisfy `1 <= size(ws.T, 1) <= min(size(A)...)`. The second dimension of `T` must equal the smallest
dimension of `A`, i.e. `size(ws.T, 2) == size(A, 2)`. If this is not the case and
`resize==true`, the workspace will automatically be resized to the appropriate dimensions.

`A` and `ws.T` are modified in-place.
"""
geqrt!(ws::QRWYWs, A::AbstractMatrix; kwargs...)

"""
    QRPivotedWs

Workspace to be used with the [`LinearAlgebra.QRPivoted`](https://docs.julialang.org/en/v1/stdlib/LinearAlgebra/#LinearAlgebra.QRPivoted)
representation of the QR factorization which uses the [`LAPACK.geqp3!`](@ref) function.

# Examples
```jldoctest
julia> A = [1.2 2.3
            6.2 3.3]
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3

julia> ws = QRPivotedWs(A)
QRPivotedWs{Float64}
  work: 100-element Vector{Float64}
  τ: 2-element Vector{Float64}
  jpvt: 2-element Vector{Int64}

julia> t = QRPivoted(LAPACK.geqp3!(ws, A)...)
QRPivoted{Float64, Matrix{Float64}, Vector{Float64}, Vector{Int64}}
Q factor:
2×2 QRPackedQ{Float64, Matrix{Float64}, Vector{Float64}}:
 -0.190022  -0.98178
 -0.98178    0.190022
R factor:
2×2 Matrix{Float64}:
 -6.31506  -3.67692
  0.0      -1.63102
permutation:
2-element Vector{Int64}:
 1
 2

julia> Matrix(t)
2×2 Matrix{Float64}:
 1.2  2.3
 6.2  3.3
```
"""
struct QRPivotedWs{T<:Number} <: Workspace
    work::Vector{T}
    τ::Vector{T}
    jpvt::Vector{BlasInt}
end

for (geqp3, elty) in ((:dgeqp3_, :Float64),
                      (:sgeqp3_, :Float32),
                      (:zgeqp3_, :ComplexF64),
                      (:cgeqp3_, :ComplexF32))
    @eval begin
        function Base.resize!(ws::QRPivotedWs, A::StridedMatrix{$elty})
            require_one_based_indexing(A)
            chkstride1(A)
            m, n = size(A)
            RldA = max(1, stride(A, 2))
            resize!(ws.jpvt, n)
            resize!(ws.τ, min(m, n))
            info = Ref{BlasInt}()
            ccall((@blasfunc($geqp3), liblapack), Cvoid,
                  (Ref{BlasInt}, Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt}, Ptr{BlasInt},
                   Ptr{$elty}, Ptr{$elty}, Ref{BlasInt}, Ref{BlasInt}),
                  m, n, A, RldA, ws.jpvt, ws.τ, ws.work, -1, info)
            chklapackerror(info[])
            resize!(ws.work, BlasInt(real(ws.work[1])))
            return ws
        end
        QRPivotedWs(A::StridedMatrix{$elty}) = 
            resize!(QRPivotedWs(Vector{$elty}(undef, 1), $elty[], BlasInt[]), A)

        function geqp3!(ws::QRPivotedWs{$elty}, A::AbstractMatrix{$elty}; resize=true)
            m, n = size(A)
            if length(ws.τ) != min(m, n)
                if resize
                    resize!(ws, A)
                else
                    throw(ArgumentError("Workspace is too small, use resize!(ws, A)."))
                end
            end
            if length(ws.jpvt) != n
                if resize
                    resize!(ws, A)
                else
                    throw(ArgumentError("Workspace is too small, use resize!(ws, A)."))
                end
            end
            lda = stride(A, 2)
            if lda == 0 # Early exit
                return A, ws.τ, ws.jpvt
            end
            info = Ref{BlasInt}()
            ccall((@blasfunc($geqp3), liblapack), Cvoid,
                  (Ref{BlasInt}, Ref{BlasInt}, Ptr{$elty}, Ref{BlasInt},
                   Ptr{BlasInt}, Ptr{$elty}, Ptr{$elty}, Ref{BlasInt},
                   Ptr{BlasInt}),
                  m, n, A, lda,
                  ws.jpvt, ws.τ, ws.work,
                  length(ws.work), info)
            chklapackerror(info[])
            return A, ws.τ, ws.jpvt
        end
    end
end

"""
    geqp3!(ws, A; resize=true) -> (A, ws.τ, ws.jpvt)

Compute the pivoted `QR` factorization of `A`, `AP = QR` using BLAS level 3,
using the preallocated [`QRPivotedWs`](@ref) workspace `ws`.
`P` is a pivoting matrix, represented by `ws.jpvt`. `ws.τ` stores the elementary
reflectors. `ws.jpvt` must have length greater
than or equal to `n` if `A` is an `(m x n)` matrix and `ws.τ` must have length
greater than or equal to the smallest dimension of `A`. If this is not the case
and `resize == true` the workspace will be appropriately resized.

`A`, `ws.jpvt`, and `ws.τ` are modified in-place.
"""
geqp3!(ws::QRPivotedWs, A::AbstractMatrix; kwargs...)
